// Koopman Checksum examples -- Chapter 7 -- unit tests

// Copyright 2024 Philip Koopman
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "support.h"
#include "chapter7_examples.h"

static const uint32_t initialSeed = SEED;  // initial sum

// Test Koopman checksum versions
void TryKoopman(uint8_t dataWord[], uint32_t dwSize)
{ 
  const uint32_t modulus8 = 253;
  const uint32_t modulus16 = 65519;
  const uint32_t modulus32 = 4294967291;

  uint32_t resultA = (uint32_t) Koopman8B(dataWord, dwSize, modulus8);
  uint32_t resultB = (uint32_t) Koopman8W(dataWord, dwSize, modulus8);
  printf("Result Koopman8B=\t0x%02X\n", resultA);
  printf("Result Koopman8W=\t0x%02X\n", resultB);
  // Make sure they both got the same result 
  assert(resultA == resultB);

  uint32_t resultC = (uint32_t) Koopman16B(dataWord, dwSize, modulus16);
  uint32_t resultD = (uint32_t) Koopman16W(dataWord, dwSize, modulus16);
  printf("Result Koopman16B=\t0x%04X\n", resultC);
  printf("Result Koopman16W=\t0x%04X\n", resultD);
  // Make sure they both got the same result 
  assert(resultC == resultD);

  uint32_t resultE = (uint32_t) Koopman32B(dataWord, dwSize, modulus32);
  uint32_t resultF = (uint32_t) Koopman32W(dataWord, dwSize, modulus32);
  printf("Result Koopman32B=\t0x%08X\n", resultE);
  printf("Result Koopman32W=\t0x%08X\n", resultF);
  // Make sure they both got the same result 
  assert(resultE == resultF);

#ifdef VALIDATE
  if(initialSeed == 0)
  {
    assert(resultA == 0x2B);
    assert(resultC == 0xFEE0);
    assert(resultE == 0xD7F61445);
  } else
  if(initialSeed == 0x9D)
  { 
    assert(resultA == 0xE5);
    assert(resultC == 0xA36C);
    assert(resultE == 0xDAF61481);
  } else
  if(initialSeed == 0xFF)
  { 
    assert(resultA == 0x9D);
    assert(resultC == 0x2644);
    assert(resultE == 0x2CF61495);
  } else
  { assert(false);
  }
#endif
}

// -------------------------------
// Run all unit tests
int main(void) 
{ 
#ifdef VALIDATE
  const uint32_t dwAlloc = 16; // divisible by 4 and >= dwSize
  uint32_t dwSize = 8;         // divisible by 4
#else
  const uint32_t dwAlloc = 24; // divisible by 4 and >= dwSize
  uint32_t dwSize = 24;        // divisible by 4
#endif
  uint8_t dataWord[dwAlloc];
  
  assert((dwAlloc & 3) == 0); // evenly divisible by 4

  InitializeData(dataWord, dwAlloc); // initialize padding bytes too
  PrintWord(dataWord, dwSize);

  printf(">>>>> Initial Seed = %d\n", initialSeed);
  TryKoopman(dataWord, dwSize);

  printf("<<<<<<<<<<<<<<<<<<<<<<<<<<< PASS >>>>>>>>>>>>>>>>>>>>>>>>>>>\n");    
  return EXIT_SUCCESS;
}

