// CRC examples -- Chapter 9 -- unit tests

// Copyright 2024 Philip Koopman
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "support.h"
#include "chapter9_examples.h"

////////////////////////////////////////////////////////
////////////////////////////////////////////////////////
// Testing code below this point

// -------------------
// Helper functions -- bit reverse

// one-byte bit reverse
uint8_t bitRevTable[256] = {
  0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
  0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 
  0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 
  0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8, 
  0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 
  0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 
  0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 
  0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC, 
  0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 
  0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2, 
  0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 
  0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA, 
  0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 
  0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6, 
  0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 
  0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE, 
  0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 
  0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1, 
  0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 
  0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9, 
  0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 
  0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5, 
  0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 
  0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD, 
  0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 
  0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 
  0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 
  0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB, 
  0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 
  0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7, 
  0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 
  0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF };

// Reverse the order of an integer
// value: bit pattern to be reversed
// bits: number of bits in pattern (1..32)
static uint32_t BitReverse(uint32_t val, uint32_t bits)
{ 
  uint32_t result = 0;
  for( uint32_t bitNum = 0; bitNum < bits; bitNum++)
  { result <<= 1;
    result |= (val & 1);
    val >>= 1;
  }
  return(result);
}

// Speed optimized bit reverse for 8-bit values
static uint8_t BitReverse8(uint8_t val)
{ 
  uint8_t result = bitRevTable[val];
  return(result);
}

// Speed optimized bit reverse for 16-bit values
static uint16_t BitReverse16(uint16_t val)
{
  const uint32_t byteMask = 0xFF;
   
  uint16_t result = bitRevTable[val&byteMask] << bitsPerByte;
  result         |= bitRevTable[(val >> bitsPerByte) & byteMask];
  return(result);
}

// Reverse the bit order in each byte of the data word
// dw[]: data word organized as 16-bit blocks
// dwSize: number of 8-bit blocks
//     Each 8-bit block is reversed in bit order
void BitReverseDW8(uint8_t dw[], uint32_t dwSize)
{
  for ( uint32_t block = 0; block < dwSize; block++)
  { dw[block] = BitReverse8(dw[block]);
  }
}

// Reverse the bit order in each block of the data word
// dw16[]: data word organized as 16-bit blocks
// dwSize: number of 16-bit blocks
//     Each 16-bit block is reversed in bit order
void BitReverseDW16(uint16_t dw16[], uint32_t dwSize)
{
  for ( uint32_t block = 0; block < dwSize ; block++)
  { dw16[block] = BitReverse16(dw16[block]);
  }
}

// -------------------

// Roll an initial seed value backwards for Naive Left shift algorithm
uint32_t Rollback16L(uint16_t initialSeed, uint16_t modulus)
{ uint16_t result = initialSeed;

  for(uint32_t iter=0; iter<16; iter++)     
  { 
   // Conditional XOR based on bottom bit of remainder
   uint16_t bottom = result & 1;
   if (bottom)
   { result ^= modulus;
     result >>= 1;
     result ^= 0x8000;
   } else
   { result >>= 1;
   }
  }
 return(result); 
}

////////////////////////////////////////////////////////
// Test functions

// Test CRC CW16 Left Shift with default polynomials
uint16_t TestCRCLeft16(uint16_t codeWordW[], uint32_t cwSize,
       uint32_t initialSeed, uint32_t finalXOR)
{
  uint16_t * checkValue = &codeWordW[cwSize-1]; // handy pointer

  // Roll back the initialSeed to account for extra iterations of naive implementation
  const uint32_t modulus16L = 0x3D65; // implicit top bit format
  uint32_t naiveSeed = Rollback16L(initialSeed, 0x3D65);
  
  printf("Initial Seed=0x%04X\tRolled back=0x%04X\tModulus=0x%04X\n", 
     initialSeed, naiveSeed, modulus16L);
  
  *checkValue = 0;        // zero and compute check value
  uint16_t resultA =  CRC16_LeftA(codeWordW, cwSize, naiveSeed, finalXOR);

  *checkValue = resultA;  // set and validate check value
  uint16_t resultVA = CRC16_LeftA(codeWordW, cwSize, naiveSeed, finalXOR);

  printf("CRC16_LeftA=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", 
     resultA, BitReverse16(resultA), resultVA);

  *checkValue = 0;        // zero and compute check value
  uint16_t resultB =  CRC16_LeftB(codeWordW, cwSize, naiveSeed, finalXOR);
  
  *checkValue = resultB;  // set and validate check value
  uint16_t resultVB = CRC16_LeftB(codeWordW, cwSize, naiveSeed, finalXOR);
  
  printf("CRC16_LeftB=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", 
     resultB, BitReverse16(resultB), resultVB);

  *checkValue = 0;        // zero and compute check value
  uint16_t resultC =  CRC16_LeftC(codeWordW, cwSize, initialSeed, finalXOR);
  
  *checkValue = resultB;  // set and validate check value
  uint16_t resultVC = CRC16_LeftC(codeWordW, cwSize, initialSeed, finalXOR); 
  
  printf("CRC16_LeftC=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", 
     resultC, BitReverse16(resultC), resultVC);

  assert(resultVA == 0);  // test adjusted CRC outcome for zero
  assert(resultVB == 0);  // test adjusted CRC outcome for zero
  assert(resultVC == 0);  // test adjusted CRC outcome for zero

  // check for all versions reaching same answer
  assert(resultA == resultB);   // same answers
  assert(resultB == resultC);   // same answers

  return(resultA);
}


// Test CRC CW16 Right Shift with default polynomials
uint16_t TestCRCRight16(uint16_t codeWordW[], uint32_t cwSize,
       uint32_t initialSeed, uint32_t finalXOR)
{
  uint16_t * checkValue = &codeWordW[cwSize-1]; // handy pointer
    
  *checkValue = 0;        // zero and compute check value
  uint16_t resultD =  CRC16_RightD(codeWordW, cwSize, initialSeed, finalXOR);
  
  *checkValue = resultD;  // set and validate check value
  uint16_t resultVD = CRC16_RightD(codeWordW, cwSize, initialSeed, finalXOR);
  
  printf("CRC16_RightD=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", resultD, BitReverse16(resultD), resultVD);

  *checkValue = 0;        // zero and compute check value
  uint16_t resultE =  CRC16_RightE(codeWordW, cwSize, initialSeed, finalXOR);
  
  *checkValue = resultE;  // set and validate check value
  uint16_t resultVE = CRC16_RightE(codeWordW, cwSize, initialSeed, finalXOR);
  
  printf("CRC16_RightE=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", resultE, BitReverse16(resultE), resultVE);

  assert(resultVD == 0);  // test adjusted CRC outcome for zero
  assert(resultVE == 0);  // test adjusted CRC outcome for zero
  
  // check for all versions reaching same answer
  assert(resultD == resultE);

  return(resultD);
}

// helper to store two-byte check value
uint16_t StoreCheckValue16(uint8_t *checkValuePtr, uint16_t value)
{
  const uint32_t byteMask = 0xFF;
  *(checkValuePtr+0) = value      & byteMask; 
  *(checkValuePtr+1) = (value>>8) & byteMask; 
  return(value);
}

uint16_t StoreCheckValueSwap16(uint8_t *checkValuePtr, uint16_t value)
{
  const uint32_t byteMask = 0xFF;
  *(checkValuePtr+1) = value      & byteMask; 
  *(checkValuePtr+0) = (value>>8) & byteMask; 
  return(value);
}

// Test CRC CW16 Left Generic (byte-organized) with default polynomials
uint16_t TestCRCGenericL(uint8_t codeWord[], uint32_t cwSize, uint32_t modulus16L, uint32_t initialSeed, uint32_t finalXOR)
{ // assumes 16-bit modulus
  const uint32_t bits = 16;
  
  uint32_t dwSize = cwSize - 2;
  uint8_t * checkValuePtr = &codeWord[dwSize]; // handy pointer

  StoreCheckValueSwap16(checkValuePtr, 0); // just to be sure
  uint16_t resultA =  CRC_LGeneric(codeWord, dwSize, modulus16L, initialSeed, finalXOR, bits);
  
  StoreCheckValueSwap16(checkValuePtr, resultA^finalXOR);
  uint16_t resultVA = CRC_LGeneric(codeWord, dwSize, modulus16L, initialSeed, finalXOR, bits);

  printf("CRC_LGeneric=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", resultA, BitReverse16(resultA), resultA^resultVA);
  

  StoreCheckValueSwap16(checkValuePtr, 0); // just to be sure
  uint16_t resultB =  CRC_LGenericT(codeWord, dwSize, modulus16L, initialSeed, finalXOR, bits);
  StoreCheckValueSwap16(checkValuePtr, resultB^finalXOR);
  uint16_t resultVB = CRC_LGenericT(codeWord, dwSize, modulus16L, initialSeed, finalXOR, bits);

  printf("CRC_LGenericT=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", resultB, BitReverse16(resultB), resultB^resultVB);

  // check for all versions reaching same answer
  assert(resultA == resultB);
  assert(resultA == resultVA);
  assert(resultB == resultVB);
  
  return(resultA);
}

// Test CRC DW16 Right Generic (byte-organized) with default polynomials
uint16_t TestCRCGenericR(uint8_t codeWord[], uint32_t cwSize, uint32_t modulus16R, uint32_t initialSeed, uint32_t finalXOR)
{ // assumes 16-bit modulus
  assert((modulus16R & 0xFFFF8000) == 0x00008000);
  
  uint32_t dwSize = cwSize - 2;
  uint8_t * checkValuePtr = &codeWord[dwSize]; // handy pointer

  StoreCheckValue16(checkValuePtr, 0); // just to be sure
  uint16_t resultA =  CRC_RGeneric(codeWord, dwSize, modulus16R, initialSeed, finalXOR);
  StoreCheckValue16(checkValuePtr, resultA^finalXOR );
  uint16_t resultVA = CRC_RGeneric(codeWord, dwSize, modulus16R, initialSeed, finalXOR);

  printf("CRC_RGeneric=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", resultA, BitReverse16(resultA),  resultVA ^ resultVA);

  StoreCheckValue16(checkValuePtr, 0); // just to be sure
  uint16_t resultB =  CRC_RGenericT(codeWord, dwSize, modulus16R, initialSeed, finalXOR);
  StoreCheckValue16(checkValuePtr, resultB^finalXOR );
  uint16_t resultVB = CRC_RGenericT(codeWord, dwSize, modulus16R, initialSeed, finalXOR);

  printf("CRC_RGenericT=\t0x%04X\t// 0x%04X\tValidation=\t0x%04X\n", resultB, BitReverse16(resultB), resultVB ^ resultVB);

  // check for all versions reaching same answer
  assert(resultA == resultB);
  assert(resultA == resultVA);
  assert(resultB == resultVB);
  
  return(resultA);
}

// Test both left and right shift algorithms on left-oriented data
uint16_t Test16(uint16_t codeWordW[], uint32_t cwSize,
       uint32_t initialSeed, uint32_t finalXOR)
{
  uint16_t result_Left16 = TestCRCLeft16(codeWordW, cwSize, initialSeed, finalXOR);
  
  // Right shift algorithm will work on reversed data and reversed result
  BitReverseDW16(codeWordW, cwSize);
  uint16_t * checkValuePtr = &codeWordW[cwSize-1]; // handy pointer
  *checkValuePtr = 0;
  
  PrintWord16(codeWordW, cwSize);
  uint16_t result_Right16 = TestCRCRight16(codeWordW, cwSize, BitReverse16(initialSeed), BitReverse16(finalXOR));
  BitReverseDW16(codeWordW, cwSize);

  assert(result_Left16 == BitReverse16(result_Right16));

#ifdef VALIDATE
  if(initialSeed == 0)
  { assert((result_Left16^finalXOR) == 0xC691);  // match manual calculation
    assert((result_Right16^BitReverse16(finalXOR)) == 0x8963); // match manual calculation
  }
#endif
  return(result_Left16);
}

uint32_t StoreCheckValue32(uint8_t *checkValuePtr, uint32_t value)
{
  const uint32_t byteMask = 0xFF;
  *(checkValuePtr+0) = value       & byteMask; 
  *(checkValuePtr+1) = (value>>8)  & byteMask; 
  *(checkValuePtr+2) = (value>>16) & byteMask; 
  *(checkValuePtr+3) = (value>>24) & byteMask;
  return(value);
}

////////////////////////////////////////////////////
// Test k-bit CRC to cover sizes 9..31 bits
// Uses byte-sized data word
void TestCRCk(uint8_t dataWord[], uint32_t dwSize, uint32_t dwAllocateSizeB, uint32_t modulusL, uint32_t bits)
{ const uint32_t initialSeed = 0x5A;
  const uint32_t finalXOR = 0x37;
  assert((bits >= 8) && (bits < 32));
  
  uint32_t modulusR = BitReverse(modulusL,bits+1) >> 1;

  // Use only unadjusted algorithms; no need to initialize check value
  InitializeData(dataWord, dwAllocateSizeB); 
  PrintWord(dataWord, dwSize);
  
  uint32_t resultA = CRC_LGeneric(dataWord, dwSize, modulusL, initialSeed, finalXOR, bits) ^ finalXOR;
  printf("CRC_LGeneric(0x%X)=\t0x%08X // 0x%08X\t", modulusL, resultA,BitReverse(resultA,bits));
  
  uint32_t resultB = CRC_LGenericT(dataWord, dwSize, modulusL, initialSeed, finalXOR, bits) ^ finalXOR;
  printf("Table: 0x%08X // 0x%08X\n", resultB,BitReverse(resultB,bits));

  BitReverseDW8(dataWord, dwAllocateSizeB);
  
  uint32_t resultC = CRC_RGeneric(dataWord, dwSize, modulusR, BitReverse(initialSeed,bits), finalXOR) ^ finalXOR;
  printf("CRC_RGeneric(0x%X)=\t0x%08X // 0x%08X\t", modulusR, resultC,BitReverse(resultC,bits));

  uint32_t resultD = CRC_RGenericT(dataWord, dwSize, modulusR, BitReverse(initialSeed,bits), finalXOR) ^ finalXOR;
  printf("Table: 0x%08X // 0x%08X\n", resultD,BitReverse(resultD,bits));

  assert(resultA == resultB);
  assert(resultC == resultD);
  assert(resultA == BitReverse(resultC,bits));
}

////////////////////////////////////////////////////
// IEEE 32-bit support

// helper to store four-byte check value
uint32_t StoreCheckValue32Swap(uint8_t *checkValuePtr, uint32_t value)
{
  const uint32_t byteMask = 0xFF;
  *(checkValuePtr+3) = value       & byteMask; 
  *(checkValuePtr+2) = (value>>8)  & byteMask; 
  *(checkValuePtr+1) = (value>>16) & byteMask; 
  *(checkValuePtr+0) = (value>>24) & byteMask; 
  return(value);
}

// Do a 32-bit CRC test for IEEE 802.3
// In Validation mode it is compared against a known check string
//     Left shift is the reference algorithm
uint32_t TestIEEELeft(uint8_t codeWord[], uint32_t cwSize)
{ 
  const uint32_t bits = 32;
  const uint32_t modulusCRC32L = 0x04C11DB7; // implicit top bit
  const uint32_t initialSeed = 0xFFFFFFFF;
  const uint32_t finalXOR = 0xFFFFFFFF;
  
  uint32_t dwSize = cwSize - 4;
  uint8_t *checkValue = &codeWord[dwSize];
  
  // initialize check value
  StoreCheckValue32Swap(checkValue, 0);
  PrintWord(codeWord, cwSize);
  
  uint32_t resultM = CRC_LGeneric(codeWord, dwSize, modulusCRC32L, initialSeed, finalXOR, bits);
  
  StoreCheckValue32Swap(checkValue, resultM ^ finalXOR);
    uint32_t resultVM = CRC_LGeneric(codeWord, dwSize, modulusCRC32L, initialSeed, finalXOR, bits);

  assert(resultM == resultVM);

  printf("CRC_LGeneric32=\t0x%08X // 0x%08X\tValidation=\t0x%08X\n", resultM,BitReverse(resultM,bits), resultM ^ resultVM);
  
  return(resultM);
}

// Do a 32-bit CRC test for IEEE 802.3 using a right-shift algorithm
//   Data and result must be bit reversed to match with left shift version
uint32_t TestIEEERight(uint8_t codeWord[], uint32_t cwSize)
{ 
  const uint32_t bits = 32;
  const uint32_t modulusCRC32R = 0xedb88320; // bit reversed; implicit +1
  const uint32_t initialSeed = 0xFFFFFFFF;
  const uint32_t finalXOR = 0xFFFFFFFF;
  
  uint32_t dwSize = cwSize - 4;
  uint8_t *checkValue = &codeWord[dwSize];
  
  // initialize check value
  StoreCheckValue32(checkValue, 0);
  PrintWord(codeWord, cwSize);

  uint32_t resultM = CRC_RGeneric(codeWord, dwSize, modulusCRC32R,   initialSeed , finalXOR);

  StoreCheckValue32(checkValue, resultM ^ finalXOR);

  uint32_t resultVM = CRC_RGeneric(codeWord, dwSize, modulusCRC32R, initialSeed, finalXOR);

  printf("CRC_RGeneric32=\t0x%08X // 0x%08X\tValidation=\t0x%08X\n", resultM, BitReverse(resultM,bits), resultM ^ resultVM);  
  
  return(BitReverse(resultM,bits));
}

// replicate IEEE 802.3 validation string
//   '123456789' ==> 0xFC891918 
//   Init=0xFFFFFFFF   final XOR=0xFFFFFFFF
void TestIEEECRC32(uint8_t codeWord[], uint32_t cwSize)
{ 
  printf("IEEE 802.3 CRC32:\n");

  uint32_t resultL = TestIEEELeft(codeWord, cwSize);

  BitReverseDW8(codeWord, cwSize);
  uint32_t resultR = TestIEEERight(codeWord, cwSize);

  assert(resultR == resultL);
#ifdef VALIDATE
  // validation string from:  https://reveng.sourceforge.io/crc-catalogue/all.htm#crc.cat.crc-32-bzip2
  assert(resultR == 0xFC891918);
#endif
}

// test 32sub8 polynomial operation with 32-bit check value computation
void Test32Sub8(uint8_t codeWord[], uint32_t dwSize, uint32_t modulus)
{ 
  uint32_t bits = 32;
  const uint32_t initialSeed = 0x1;
  const uint32_t finalXOR = 0;

  uint32_t resultA = CRC_RGenericT32sub8(codeWord, dwSize, 
                 modulus, initialSeed, finalXOR);
      
  printf("CRC_Test32Sub8R(0x%08X)=\t0x%08X // 0x%08X\t", modulus, resultA, BitReverse(resultA,bits));        

  uint32_t resultB = CRC_RGeneric(codeWord, dwSize, 
                 modulus, initialSeed, finalXOR);
      
  printf("CRC_TestR=\t0x%08X // 0x%08X\n", resultB, BitReverse(resultB,bits));        
  assert (resultA == resultB);               

}

// test 64sub8 polynomial operation with 32-bit check value computation
void Test64Sub8(uint8_t codeWord[], uint32_t dwSize, uint64_t modulus)
{ 
  const uint64_t initialSeed = 0x1;
  const uint64_t finalXOR = 0;

  uint64_t resultA = CRC_RGenericT64sub8(codeWord, dwSize, 
                 modulus, initialSeed, finalXOR);
  printf("CRC_Test64Sub8R(0x%08lX)=\t0x%08lX\t", modulus, resultA);        
  
  uint64_t resultB = CRC_RGeneric64(codeWord, dwSize, 
                 modulus, initialSeed, finalXOR);
  printf("CRC_Test64R=\t0x%08lX\n", resultB);        
  
  assert (resultA == resultB);               
}


// ----------------------
// Run all unit tests
int main(void) 
{ 
// only print out CRC lookup table if compiled with #define TABLE
#if TABLE
 crcLookupPrintR(0xedb88320);
 return(EXIT_SUCCESS);
#endif 
 
  // Fixed test moduli
  // Style note: Duplicated in each procedure
  //   to optimize for textbook self-contained examples.
  //   Otherwise these should only be defined in one place.
  const uint16_t modulus16R = 0xA6BC;
  const uint32_t modulus16L = 0x13D65; // explicit top bit format
  
#ifdef VALIDATE
  // define code word and data word sizes in blocks
  const uint16_t initialSeed = 0;      // to match hand-done values
  const uint16_t finalXOR = 0x5678;    // can be zero if desired
  const uint32_t cwAllocateSizeW = 5; 
  uint32_t cwSizeW = 5;       
  const uint32_t cwAllocateSizeB = 16; // divisible by 4 and >= cwSize
  uint32_t cwSizeB = 10;               // divisible by 4
#else
  // define code word and data word sizes in blocks for larger examples
  const uint16_t initialSeed = 0xFFFF; // can be zero if desired
  const uint16_t finalXOR = 0xABCD;    // can be zero if desired
  const uint32_t cwAllocateSizeW = 13;
  uint32_t cwSizeW = 13;       
  const uint32_t cwAllocateSizeB = 32; // divisible by 4 and >= cwSize
  uint32_t cwSizeB = 26;               // divisible by 4
#endif

  // Set up data code word arrays for 16-bit blocks and 8-bit blocks
  uint16_t codeWordW[cwAllocateSizeW];
  uint8_t codeWord[cwAllocateSizeB];
  InitializeData16(codeWord, cwAllocateSizeB, codeWordW, cwSizeW);
  
  printf("//// testing 16 bit blocks; cwSize=%d blocks\n", cwSizeW);
  codeWordW[cwSizeW-1] = 0;     // zero check value
  PrintWord16(codeWordW, cwSizeW);
  
  uint32_t result_16 = Test16(codeWordW, cwSizeW, initialSeed, finalXOR);

  printf("//// testing 8 bit blocks; cwSize=%d blocks\n", cwSizeB);
  StoreCheckValue16(codeWord+cwSizeB-2, 0); // zero check value
  PrintWord(codeWord, cwSizeB);
  
  uint32_t resultG = TestCRCGenericL(codeWord, cwSizeB, modulus16L, initialSeed, finalXOR);  
  BitReverseDW8(codeWord, cwSizeB);
  uint32_t resultH = TestCRCGenericR(codeWord, cwSizeB, modulus16R, BitReverse16(initialSeed), BitReverse16(finalXOR));
  
  assert((resultG) == BitReverse((resultH),16));
  assert(resultG == result_16);

#ifdef VALIDATE
  if(initialSeed == 0)
  { assert((resultG^finalXOR) == 0xC691); // manual computation
  }
#endif

 // Test other size CRC polynomials
 uint32_t dwSizeB = cwSizeB-1;
 TestCRCk(codeWord, dwSizeB, cwAllocateSizeB, 0x1CF, 8);
 TestCRCk(codeWord, dwSizeB, cwAllocateSizeB, 0x233, 9);
 TestCRCk(codeWord, dwSizeB, cwAllocateSizeB, 0x9E47, 15);
 TestCRCk(codeWord, dwSizeB, cwAllocateSizeB, 0x2df4f, 17);
 TestCRCk(codeWord, dwSizeB, cwAllocateSizeB, 0xa81be1, 23);
 TestCRCk(codeWord, dwSizeB, cwAllocateSizeB, 0xe6fc6257, 31);
   
 // Example lookup table values for narrative description in book
 // printf("crcLookup16[%04X] = %04X\n", 0x536E, crcLookup16[0x536E] );
 // printf("crcLookupTableR[%02X] = %04X\n", 0xB3, crcLookupTableR[0xB3] );

 Test32Sub8(codeWord, dwSizeB, 0xA7000000);
 Test32Sub8(codeWord, dwSizeB, 0x4F000000);
 Test32Sub8(codeWord, dwSizeB, 0x2F000000);
 Test32Sub8(codeWord, dwSizeB, 0x1F000000);
 Test32Sub8(codeWord, dwSizeB, 0x0F000000);
 Test32Sub8(codeWord, dwSizeB, 0x07000000);
 Test32Sub8(codeWord, dwSizeB, 0x03000000);
 
 Test64Sub8(codeWord, dwSizeB, 0xD800000000000000UL);

 cwSizeB = 13;
 InitializeData(codeWord, cwAllocateSizeB); 
 TestIEEECRC32(codeWord, cwSizeB);
 
  
 printf("<<<<<<<<<<<<<<<<<<<<<<<<<<< PASS >>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
 return EXIT_SUCCESS;
}

